@@ -1,4 +1,6 @@
+#include <algorithm>
 #include <hyprtoolkit/core/Backend.hpp>
+#include <hyprtoolkit/core/Timer.hpp>
 #include <hyprtoolkit/palette/Palette.hpp>
 
 #include "InternalBackend.hpp"
@@ -9,7 +11,6 @@
 #include "../window/WaylandWindow.hpp"
 #include "../window/WaylandLayer.hpp"
 #include "../Macros.hpp"
-#include "../helpers/Timer.hpp"
 #include "../element/Element.hpp"
 #include "../palette/ConfigManager.hpp"
 #include "../system/Icons.hpp"
@@ -54,7 +55,7 @@ CBackend::~CBackend() {
     close(m_sLoopState.exitfd[1]);
 }
 
-SP<CBackend> CBackend::create() {
+SP<IBackend> IBackend::create() {
     if (g_backend)
         return nullptr;
     g_backend = SP<CBackend>(new CBackend());
@@ -197,6 +198,18 @@ void CBackend::removeFd(int fd) {
     rebuildPollfds();
 }
 
+void CBackend::doOnReadable(Hyprutils::OS::CFileDescriptor fd, std::function<void()>&& fn) {
+    int fdInt = fd.get();
+    m_sLoopState.userFds.emplace_back(SFDListener{
+        .fdOwned      = std::move(fd),
+        .fd           = fdInt,
+        .callback     = std::move(fn),
+        .removeOnFire = true,
+    });
+
+    rebuildPollfds();
+}
+
 constexpr size_t INTERNAL_FDS = 3;
 
 void             CBackend::rebuildPollfds() {
@@ -375,12 +388,22 @@ void CBackend::enterLoop() {
         }
 
         // do user fds
+        std::vector<int> expiredFds;
+
         for (auto& uf : m_sLoopState.userFds) {
             if (!uf.needsDispatch)
                 continue;
 
             uf.needsDispatch = false;
             uf.callback();
+
+            if (uf.removeOnFire)
+                expiredFds.emplace_back(uf.fd);
+        }
+
+        if (!expiredFds.empty()) {
+            std::erase_if(m_sLoopState.userFds, [&expiredFds](const auto& e) { return std::ranges::contains(expiredFds, e.fd); });
+            rebuildPollfds();
         }
     }
 
